﻿using Buzz.MachineInterface;
using BuzzGUI.Common;
using BuzzGUI.Interfaces;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Globalization;
using System.Linq;
using System.Windows;
using System.Windows.Controls;

namespace WDE.TimeLine
{
	[MachineDecl(Name = "TimeLine", ShortName = "TimeLine", Author = "WDE", MaxTracks = 1)]
	public class TimeLine : IBuzzMachine, INotifyPropertyChanged
	{
		public IBuzzMachineHost host;

		public TimeLine(IBuzzMachineHost host)
		{
			this.host = host;			
		}

		[ParameterDecl(ValueDescriptions = new[] { "no", "yes" })]
		public bool Bypass { get; set; }
	
		// actual machine ends here. the stuff below demonstrates some other features of the api.
	
		public class State : INotifyPropertyChanged
		{
			public State() { text = "here is state"; }	// NOTE: parameterless constructor is required by the xml serializer

			string text;
			public string Text 
			{
				get { return text; }
				set
				{
					text = value;
					if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs("Text"));
					// NOTE: the INotifyPropertyChanged stuff is only used for data binding in the GUI in this demo. it is not required by the serializer.
				}
			}

			public event PropertyChangedEventHandler PropertyChanged;
		}

		State machineState = new State();
		public State MachineState			// a property called 'MachineState' gets automatically saved in songs and presets
		{
			get { return machineState; }
			set
			{
				machineState = value;
				if (PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs("MachineState"));
			}
		}		
		
		int checkedItem = 1;
		
		public IEnumerable<IMenuItem> Commands
		{
			get
			{
				yield return new MenuItemVM() 
				{ 
					Text = "About...", 
					Command = new SimpleCommand()
					{
						CanExecuteDelegate = p => true,
						ExecuteDelegate = p => MessageBox.Show("WDE TimeLine 0.1")
					}
				};
			}
		}		

		public event PropertyChangedEventHandler PropertyChanged;

		internal void UpdateTimeLine(int timeLineSizeInTicks)
		{
			int songLength = host.Machine.Graph.Buzz.Song.SongEnd;
			int position = 0;

			int beatsPerMin = host.MasterInfo.BeatsPerMin;
			int tickPerBeat = host.MasterInfo.TicksPerBeat;
			double sPerTick = 1.0 / (beatsPerMin / 60.0 * tickPerBeat);

			DateTime dt = new DateTime(0);
			
			DeleteAllPatterns();

			ISequence seq = GetSequence(); // First one

			while (position < songLength)
			{
				string name = dt.ToString("m:ss:ff", DateTimeFormatInfo.InvariantInfo);
				host.Machine.CreatePattern(name, timeLineSizeInTicks);

				IPattern pat = GetPattern(name);
				SequenceEvent seqe = new SequenceEvent(SequenceEventType.PlayPattern, pat);
								
				seq.SetEvent(position, seqe);

				double secondsToAdd = CalculateTimeLinePatternLenghtInSeconds(position, timeLineSizeInTicks, ref beatsPerMin, ref tickPerBeat);
				// double secondsToAdd = sPerTick * timeLineSizeInTicks;

				dt = dt.AddSeconds( secondsToAdd );

				position += timeLineSizeInTicks;
			}
		}

		private double CalculateTimeLinePatternLenghtInSeconds(int position, int timeLineSizeInTicks, ref int beatsPerMin, ref int tickPerBeat)
		{
			double totalTimeInSeconds = 0;			

			for (int movingPosition = position; movingPosition < position + timeLineSizeInTicks; movingPosition++)
			{				
				foreach (ISequence seq in host.Machine.Graph.Buzz.Song.Sequences)
				{
					if (seq.Machine.Name == "Master") // Need to do this if multiple Master machine sequences...
					{						
						foreach (KeyValuePair<int, SequenceEvent> eve in seq.Events)
						{
							if (movingPosition >= eve.Key && movingPosition < eve.Key + eve.Value.Span)
							{
								UpdateBPMAndTPB(eve.Value.Pattern, movingPosition - eve.Key, ref beatsPerMin, ref tickPerBeat);
							}
						}
					}
				}

				totalTimeInSeconds += 1.0 / (beatsPerMin / 60.0 * tickPerBeat);
			}

			return totalTimeInSeconds;
		}

		private void UpdateBPMAndTPB(IPattern pat, int positionInPattern, ref int beatsPerMin, ref int tickPerBeat)
		{
			IEnumerable<PatternEvent> eventsBPM = pat.Columns[1].GetEvents(positionInPattern, positionInPattern + 1);
			IEnumerable<PatternEvent> eventsTPB = pat.Columns[2].GetEvents(positionInPattern, positionInPattern + 1);

			if (eventsBPM.Count() > 0)
			{
				beatsPerMin = eventsBPM.First().Value;
			}

			if (eventsTPB.Count() > 0)
			{
				tickPerBeat = eventsTPB.First().Value;
			}			
		}

		private IPattern GetPattern(string name)
		{
			IPattern ret = null;
			foreach (IPattern pat in host.Machine.Patterns)
			{
				if (pat.Name == name)
					return pat;
			}
			return ret;
		}

		private ISequence GetSequence()
		{
			foreach (ISequence seq in host.Machine.Graph.Buzz.Song.Sequences)
			{
				if (seq.Machine == host.Machine)
					return seq;
			}
			return null;
		}

		private void DeleteAllPatterns()
		{			
			while (host.Machine.Patterns.Count > 0)
			{
				IPattern pat = host.Machine.Patterns.First();
				host.Machine.DeletePattern(pat);
			}			
		}
	}

	public class MachineGUIFactory : IMachineGUIFactory { public IMachineGUI CreateGUI(IMachineGUIHost host) { return new TimeLineGUI(); } }
	public class TimeLineGUI : UserControl, IMachineGUI
	{
		IMachine machine;
		TimeLine timeLineMachine;
		public NumericUpDown numBoxTicks;

		// view model for machine list box items
		public class MachineVM
		{
			public IMachine Machine { get; private set; }
			public MachineVM(IMachine m) { Machine = m; }
			public override string ToString() { return Machine.Name; }
		}
		
		public IMachine Machine
		{
			get { return machine; }
			set
			{
				machine = value;

				if (machine != null)
				{
					timeLineMachine = (TimeLine)machine.ManagedMachine;
				}
			}
		}
		
		public TimeLineGUI()
		{
			numBoxTicks = new NumericUpDown() { HorizontalAlignment=HorizontalAlignment.Stretch, Height = 20, DecimalPlaces = 0, Change = (decimal)1, Minimum = (decimal)32, Maximum = (decimal)256, Margin = new Thickness(4, 4, 4, 4) };

			Button bt = new Button() { Height=26, Content = "Update", Margin = new Thickness(4, 4, 4, 4), HorizontalAlignment = HorizontalAlignment.Stretch /*, Width = 200*/ };
			bt.Click += Bt_Click;
			
			var sp = new StackPanel() { };			
			sp.Children.Add(numBoxTicks);
			sp.Children.Add(bt);
			this.Content = sp;

			this.Loaded += TimeLineGUI_Loaded;
		}

		private void TimeLineGUI_Loaded(object sender, RoutedEventArgs e)
		{
			timeLineMachine.host.Machine.ParameterWindow.ResizeMode = ResizeMode.NoResize;

			Grid content = (Grid)timeLineMachine.host.Machine.ParameterWindow.Content;
			var content2 = (content.Children[4]); // Super hack
			content2.Visibility = Visibility.Hidden;
		}

		private void Bt_Click(object sender, RoutedEventArgs e)
		{
			if (timeLineMachine != null)
			{
				timeLineMachine.UpdateTimeLine((int)numBoxTicks.Value);
			}
		}
	}
}
